#pragma comment(lib, "Bass64.lib")
#include "bass.h"

const int STREAM = 0;
const int MUSIC  = 1;

struct BASSPROP {
    WCHAR    audiofile[MAX_PATH];
    HSTREAM  channel;       // The playing audio channel
    WORD     lpeak;         // The left audio channel peak value (while in play mode).
    WORD     rpeak;         // The right audio channel peak value (while in play mode).
    LONG_PTR fftdata;       // pointer to the FFT() AS SINGLE array.
    LONG_PTR fftsize;       // Size of the FFT array.
    LONG_PTR wimdata;       // DWORD pointer to the wave MM_WIM_DATA.
    DWORD    medialength;   // Media length.
    DWORD    mediapos;      // Media pos.
    float    progress;      // Media progress.

    long     mediatype;     // MUSIC, STREAM, etc.
    long     initdone;      // BassInit flag.

    BOOL     bPause;
};

BASSPROP gB;

static const void* BassChannelGetData() {
    static float fft[256];
    ClearMemory(&fft[0], sizeof(float)* 256);
    if (gB.channel) { DWORD dwRet = BASS_ChannelGetData(gB.channel, &fft[0], BASS_DATA_FFT512); }
    return &fft[0];
}

static long BassFree() {
    long nRet = 0;
    if (gB.initdone) { gB.initdone = 0; nRet = BASS_Free(); }
    return nRet;
}

static long BassCloseChannel() {
    long nRet = 0;
    if (gB.channel) {
        if (gB.mediatype == MUSIC) {
            nRet = BASS_MusicFree(gB.channel);
//WCHAR zAf[MAX_PATH] = {0};
//swprintf(zAf, strSize(zAf), L"BASS_MusicFree (%d), errorCode %d", nRet, BASS_ErrorGetCode());
//zTrace(zAf);
        } else {
            nRet = BASS_StreamFree(gB.channel);
        }
        gB.channel = 0; gB.mediatype = 0; gB.mediapos = 0; gB.medialength = 0;
    }
    return nRet;
}

static long BassInit(IN HWND hWnd) {
    int device = -1; // -1 means use default device
    if (gB.initdone == 0) { gB.initdone = BASS_Init(device, 44100, 0, hWnd, 0); } // -1 = default device
    return gB.initdone;
}

BOOL BassSetChannelVolume(IN float percent) {
    BOOL bRet = FALSE;
    if (gB.channel) {
        BOOL bRet = BASS_ChannelSetAttribute(gB.channel, BASS_ATTRIB_VOL, percent);
    }
    return bRet;
}

long BassChannelPlay(IN HWND hWnd);
static void CALLBACK BassStreamEllapsed(IN HSYNC handle, IN DWORD channel, IN DWORD data, void *user) {
    BassChannelPlay(gP.hMain);
}

static void BassSetAudioVolume() {
    if (gB.channel) BASS_ChannelSetAttribute(gB.channel, BASS_ATTRIB_VOL, (float)(gD.setAudioLevel * 0.01f));
}

static long BassChannelPlay(IN HWND hWnd) {
    long nRet = 0;

    BassCloseChannel();

    if (BassInit(hWnd)) {
        if (MediaCheckExtension(gB.audiofile) == MEDIA_BASS_MUSIC) {
            gB.mediatype = MUSIC;
            gB.channel = BASS_MusicLoad(0, gB.audiofile, 0, 0, BASS_STREAM_PRESCAN | BASS_STREAM_AUTOFREE, 0);
        } else { // MEDIA_BASS_STREAM
            gB.mediatype = STREAM;
            gB.channel = BASS_StreamCreateFile(0, gB.audiofile, 0, 0, BASS_STREAM_AUTOFREE);
        }

        if (gB.channel) {
            gB.medialength = (DWORD) BASS_ChannelGetLength(gB.channel, 0);
            BASS_ChannelSetSync(gB.channel, BASS_SYNC_END | BASS_SYNC_ONETIME, 0, &BassStreamEllapsed, 0);
            BassSetAudioVolume();
            nRet = BASS_ChannelPlay(gB.channel, 0);
            HSTREAM nScanChannel = BASS_StreamCreateFile(0, gB.audiofile, 0, 0, BASS_STREAM_DECODE | BASS_UNICODE);
        } else {
            BassCloseChannel();
        }
    }
    return nRet;
}

static WORD SolvePeak(IN WORD nValue, IN WORD nTotal) {
    WORD nRet = 0;
    if (nTotal) nRet = (nValue * 100) / nTotal;
    return nRet;
}

static LONG_PTR BassChannelGetWimData (IN HSTREAM nChannel) {
    static float fft[256];
    ClearMemory(&fft[0], sizeof(float) * 256);
    if (nChannel) { long nRet = BASS_ChannelGetData(nChannel, &fft[0], 1024); }
    return (LONG_PTR) &fft[0];
}

DWORD BassChannelGetLevel(IN HSTREAM nChannel) {
    DWORD nLevel = 0;
    if (nChannel) { nLevel = BASS_ChannelGetLevel(nChannel); }
    return nLevel;
}

static BOOL BassChannelResume() {
    BOOL bRet = FALSE;
    if (gB.channel) {
        bRet = BASS_ChannelPlay(gB.channel, 0);
        ZD_SetObjectVisibility(ID_FLARE, ZS_VISIBLE);
    }
    gB.bPause = FALSE;
    return bRet;
}

static void BassChannelPause() {
    if (gB.channel) {
        BASS_ChannelPause(gB.channel);
        ZD_SetObjectVisibility(ID_FLARE, ZS_HIDDEN);
        gB.bPause = TRUE;
    }
}