 //+--------------------------------------------------------------------------+
 //|                                                                          |
 //|                                Plugin64                                  |
 //|                                                                          |
 //|       How to ceate a mini demoscene using Bass.dll and BBplugin.dll.     |
 //|                                                                          |
 //|                                                                          |
 //+--------------------------------------------------------------------------+
 //|                                                                          |
 //|                         Author Patrice TERRIER                           |
 //|                           copyright (c) 2024                             |
 //|                                                                          |
 //|                        pterrier@zapsolution.com                          |
 //|                           www.objreader.com                              |
 //|                          objreader@gmail.com                             |
 //|                                                                          |
 //+--------------------------------------------------------------------------+
 //|                  Project started on : 08-28-2024 (MM-DD-YYYY)            |
 //|                        Last revised : 09-20-2024 (MM-DD-YYYY)            |
 //+--------------------------------------------------------------------------+

#pragma warning(disable:4312)
//#pragma warning(disable:4302)

#include <windows.h>
#include <gl/GL.h>
#include <cwchar>  // For swprintf

#include "Tools.h"
#include "GDImage.h"
#include "WinLIFT.h"
#include "Globals.h"
#include "BassAudio.h"
#include "bbp.h"

#pragma comment(lib, "GDImage64.lib")
#pragma comment(lib, "WinLIFT64.lib")

static void gl_DrawScene() {
    if (gB.bPause == TRUE) return;

    short* pInt = (short*)BassChannelGetWimData(gB.channel);
    DWORD nLevel = BassChannelGetLevel(gB.channel);

    if (IsWindow(gP.hGL)) {
        if (IsWindowVisible(gP.hGL)) BBP_RenderOpenGL(nLevel, pInt);
    }
}

static void PaintMain(IN HWND hWin) {
    PAINTSTRUCT ps; ClearMemory(&ps, sizeof(ps));
    BeginPaint(hWin, &ps);
    EndPaint(hWin, &ps);
    gl_DrawScene();
}

// Move the transparent toolwindow
static void MoveOverlay(IN UINT uMsg) {
    RECT lpr; GetWindowRect(gP.hMain, &lpr);
    lpr.left += skGetSystemMetrics(SK_CXFRAMELEFT);
    lpr.top += skGetSystemMetrics(SK_CYFRAMETOP);
    lpr.right -= skGetSystemMetrics(SK_CXFRAMERIGHT);
    lpr.bottom -= skGetSystemMetrics(SK_CYFRAMEBOTTOM);
    if (uMsg == WM_SIZE) {
        // Update Arrow X,Y to match the new size
        long y = 0, w = 0, h = 0;
        ZD_GetObjectBound (ID_LARROW, w, h);
        y = (rHeight(lpr) - h) / 2;
        ZD_SetObjectXY(ID_LARROW, -w + 3, y, FALSE);
        ZD_SetObjectXY(ID_RARROW, rWidth(lpr) - 3, y, FALSE);
    }
    MoveWindow(gP.hGDImage, lpr.left, lpr.top, rWidth(lpr), rHeight(lpr), TRUE);
}

// Move the copyright marquee baner.
static void MoveMarquee() {
    long x = 0, y = 0, w = 0, h = 0, nID = 0;
    RECT rc; GetClientRect(gP.hGDImage, &rc);

    // Move Seek thumb
    nID = ID_SEEKTHUMB;
    ZD_GetObjectXY(nID, x, y);
    x = (long)(gB.progress * rWidth(rc)) - 7;
    ZD_SetObjectXY(nID, x, y, ZD_DRAW_DEFERRED);

    nID = ID_COPYRIGHT;
    ZD_GetObjectBound(nID, w, h);
    if (w > 0) {
        ZD_GetObjectXY(nID, x, y);
        x -= 1; if (x < -w) x = rWidth(rc);
        ZD_SetObjectXY(nID, x, y, TRUE);
    }

}

// Update the circular progress bar
static void UpdateCircularProgressBar() {
    gB.mediapos = (DWORD) BASS_ChannelGetPosition(gB.channel, 0);
    if (gB.medialength) {
        gB.progress = (float)(gB.mediapos / (float)gB.medialength);
        if (gB.progress > 0.995f) {         // Because some music tracker never stop
            BassStreamEllapsed(0, 0, 0, 0); // Loop mode
            return;
        }
    } else {
        gB.progress = 0.0f;
    }

    WCHAR zAf[32]; ClearMemory(zAf, sizeof(zAf));
    swprintf(zAf, strSize(zAf), L"%d%%", (long)(gB.progress * 100));
    ZD_SetObjectTextEx(ID_PIE_TEXT, zAf, ZD_DRAW_DEFERRED);
    ZOBJECT zObj; ClearMemory(&zObj, sizeof(zObj));
    ZD_GetObjectProperties(ID_PIE_OUT, zObj);
    zObj.sweepAngle = gB.progress * 360.0f;
    ZD_SetObjectProperties(ID_PIE_OUT, zObj);
}

static WCHAR* EXEconfig() {
    static WCHAR zPath[MAX_PATH];
    static long done;
    if (done == 0) {
        ClearMemory(zPath, sizeof(zPath));
        Path_Combine(zPath, zGetTempPath(), L"zplugin.bin");
        done = -1;
    }
    return zPath;
}

static void LoadConfig() {
    long OK = 0;
    DWORD dwBytes = 0;
    ClearMemory(&gP.bin, sizeof(gP.bin));
    gD.setAudioLevel = 50;
    if (FileExist(EXEconfig())) {
        if (FileSize(EXEconfig()) == sizeof(gP.bin)) {
            HANDLE hFile = CreateFile(EXEconfig(), GENERIC_READ, 0, NULL, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
            if (hFile != INVALID_HANDLE_VALUE) {
                if (ReadFile(hFile, &gP.bin, sizeof(gP.bin), &dwBytes, NULL)) {
                    OK = -1;
                    gD.setAudioLevel = gP.bin.audiolevel;
                }
                CloseHandle(hFile);
            }
        }
    }
}

static void SaveConfig() {
    DWORD dwBytes = 0;
    HANDLE hFile = CreateFile(EXEconfig(), GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
    if (hFile != INVALID_HANDLE_VALUE) {
        WriteFile(hFile, &gP.bin, sizeof(gP.bin), &dwBytes, NULL);
        CloseHandle(hFile);
    }
}

static LRESULT CALLBACK WndProc(IN HWND hWin, IN UINT uMsg, IN WPARAM wParam, IN LPARAM lParam) {

    MINMAXINFO* pMM;

    switch (uMsg) {

    case WM_DROPFILES:
    {
        WCHAR szFile[MAX_PATH]; ClearMemory(szFile, sizeof(szFile));
        if (DragQueryFile((HDROP) wParam, 0xFFFFFFFF, NULL, 0) > 0) {
            if (DragQueryFile((HDROP)wParam, 0, szFile, MAX_PATH)) {
                if (MediaCheckExtension(szFile) >= MEDIA_BASS_MUSIC) {
                    wcscopy(gB.audiofile, szFile);
                    MarqueeUpdate();
                    BassChannelPlay(hWin);

                    if (gB.channel) {

                        // If gB.bPause, then switch back to full playing mode
                        if (gB.bPause) {

                            ZI_SetMouseOverObjectID(ID_BTN_PAUSE);
                            ZD_SetObjectVisibility(ID_BTN_PAUSE, ZS_VISIBLE);
                            ZD_SetObjectFrameToUse(ID_BTN_PAUSE, 3, TRUE);
                            gD.overID = ID_BTN_PAUSE;

                            ZD_SetObjectFrameToUse(ID_BTN_PLAY, 1, FALSE);
                            ZD_ShowObject(ID_BTN_PLAY, ZS_HIDDEN);
                            gB.bPause = FALSE;

                        }
                        BBP_NewSound();
                    }

                }
            }
        }
        DragFinish((HDROP)wParam);

        // Use foreground window to restore correctly the caption color.
        SetForegroundWindow(hWin);
        return 0;
    }
        break;

    case WM_GETMINMAXINFO:
        pMM = (MINMAXINFO*) lParam;
        pMM->ptMinTrackSize.x = gP.MinTrackSizeW;
        pMM->ptMinTrackSize.y = gP.MinTrackSizeH;
        break;

    case WM_TIMER:
        UpdateCircularProgressBar();
        MoveMarquee();
        gl_DrawScene();
        break;

    case WM_MOUSEWHEEL:
        if (GetFocus() != gP.hGL) SetFocus(gP.hGL);
        break;

    case WM_MOVE:
    case WM_SIZE:
        MoveOverlay(uMsg);
        if (uMsg == WM_SIZE) {
            BBP_ResizeOpenGL();
            gl_DrawScene();
        }
        break;

    case WM_CLOSE:
        BassCloseChannel();
        break;

    case WM_PAINT:
        PaintMain(hWin);
        return 0;

    case WM_DESTROY:
        if (gP.hLib) {
            SaveConfig();
            BBP_Detached(hWin, gP.hLib);
        }
        BassFree();
        PostQuitMessage(0);
        return 0;
    }
    return DefWindowProc(hWin, uMsg, wParam, lParam);
}

static void CreateSliderHorz(IN HBITMAP hBitmap, IN HBITMAP hThumb, IN long nID_BACK, IN long nID_THUMB, IN long nID_PROGRESS, IN long nX, IN long nY) {
    long x, y, w, h, bmW, bmH, nID;
    ZI_GetBitmapSize(hBitmap, bmW, bmH);
    x = nX;
    y = nY;
    h = bmH;
    nID = nID_BACK;
    ZD_DrawBitmapToCtrl(gP.hGDImage, x, y, hBitmap, OPAQUEIMAGE, nID, ZS_VISIBLE);
    ZD_SetObjectAnchorMode(nID, ANCHOR_CENTER_HORZ_BOTTOM);
    ZD_SetObjectLocked(nID, TRUE);
    x = nX + (bmW / 2);

    ZI_GetBitmapSize(hThumb, bmW, bmH);
    x -= bmW / 2 - 1;
    y = nY;
    w = x - nX;
    nID = nID_PROGRESS;
    ZD_DrawRectangleToCtrl(gP.hGDImage, nX, nY + 16, w + (bmW / 2), 5, HIGHLIGHT, 0, nID, ZS_VISIBLE, ZD_DRAW_FILLED, 0);
    ZD_SetObjectAnchorMode(nID, ANCHOR_CENTER_HORZ_BOTTOM);
    ZD_SetObjectHidden(nID, 2); // 2 = Transparent, its ID won't be detected and it couldn't be moved.

    nID = nID_THUMB;
    WCHAR szFile[MAX_PATH]; ClearMemory(szFile, sizeof(szFile));
    ZD_DrawBitmapToCtrl(gP.hGDImage, x, y, hThumb, OPAQUEIMAGE, nID, ZS_VISIBLE);
    ZD_SetObjectAnchorMode(nID, ANCHOR_CENTER_HORZ_BOTTOM);
    ZD_SetObjectLocked(nID, TRUE);
    ZD_SetObjectHidden(nID, 2); // 2 = Transparent, its ID won't be detected and it couldn't be moved.
}

static BOOL IsSLIDER_HORZ(IN long nID) {
    BOOL bRet = FALSE;
    if ((nID > ID_SLIDER_HORZ - 1) && (nID < ID_THUMB_HORZ + (ID_THUMB_HORZ - ID_SLIDER_HORZ))) bRet = TRUE;
    return bRet;
}

static BOOL IsPUSHBUTTON(IN long nID) {
    BOOL bRet = FALSE;
    if ((nID > ID_PUSH_BUTTON) && (nID < ID_PUSH_BUTTON + 100)) bRet = TRUE;
    return bRet;
}

static void ProcessMessage(IN UINT uMsg, IN WPARAM wParam, IN LPARAM lParam) {
    long nID, wmID, wmEvent;

    //SetForegroundWindow(gP.hMain);

    switch (uMsg) {
    case WM_HSCROLL:
         if (LOINT(wParam) == SB_THUMBPOSITION) {
            MoveMemory(&nID, &lParam, 4);
            gD.setAudioLevel = HIINT(wParam);
            BassSetAudioVolume();
            //if (gB.channel) BASS_ChannelSetAttribute(gB.channel, BASS_ATTRIB_VOL, (float)(nPercent * 0.01f));
         }
         break;
    case WM_COMMAND:
        wmID    = LOINT(wParam);
        wmEvent = HIINT(wParam);
        switch (wmID) {
        case ID_BTN_PAUSE:
            ZI_SetMouseOverObjectID(ID_BTN_PLAY);
            ZD_SetObjectVisibility(ID_BTN_PLAY, ZS_VISIBLE);
            ZD_SetObjectFrameToUse(ID_BTN_PLAY, 3, TRUE);
            gD.overID = ID_BTN_PLAY;

            ZD_SetObjectFrameToUse(ID_BTN_PAUSE, 1, FALSE);
            ZD_ShowObject(ID_BTN_PAUSE, ZS_HIDDEN);
            BassChannelPause();
            break;
        case ID_BTN_PLAY:
            ZI_SetMouseOverObjectID(ID_BTN_PAUSE);
            ZD_SetObjectVisibility(ID_BTN_PAUSE, ZS_VISIBLE);
            ZD_SetObjectFrameToUse(ID_BTN_PAUSE, 3, TRUE);
            gD.overID = ID_BTN_PAUSE;

            ZD_SetObjectFrameToUse(ID_BTN_PLAY, 1, FALSE);
            ZD_ShowObject(ID_BTN_PLAY, ZS_HIDDEN);
            gB.bPause = BASS_ChannelIsActive(gB.channel);
            if (gB.bPause) {
                BassChannelResume();
            } else {
                BassChannelPlay(gP.hMain);
            }
            break;
        default:
            break;
        }
        break;
    default:
        break;
    }
}

static void MoveThumbHorz(IN long nID, IN long nMouseX, IN long THUMB_HORZ, IN long SLIDER_HORZ, IN long PROGRESS_HORZ) {
    long x, y, nX, nY, w, h, nW, nHalf, nPosX, nRedraw;
    if (nID < THUMB_HORZ) nID += THUMB_HORZ - SLIDER_HORZ;
    ZD_GetObjectXY(nID - (THUMB_HORZ - SLIDER_HORZ), nX, y);
    ZD_GetObjectBound(nID - (THUMB_HORZ - SLIDER_HORZ), nW, h);
    ZD_GetObjectBound(nID, w, h);
    ZD_GetObjectXY(nID, x, y); nRedraw = x;
    nHalf = w / 2;
    x = min(max(nMouseX - nHalf, nX + 3 - nHalf), nX + nW - 2 - nHalf);
    if (x != nRedraw) {
        nPosX = (nX + nW - 2) - (x + nHalf);
        gD.setAudioLevel = 100 - (long)((nPosX / (float)(nW - 5)) * 100);
        if (gD.setAudioLevel != gP.bin.audiolevel) {
            ZD_GetObjectBound(PROGRESS_HORZ, w, h);
            ZD_GetObjectXY(PROGRESS_HORZ, nX, nY);
            ZD_DrawRectangleToCtrl(gP.hGDImage, nX + 1, nY + 1, x + nHalf - (nX + 1), h - 1, HIGHLIGHT, 0, PROGRESS_HORZ, ZS_VISIBLE, ZD_DRAW_FILLED, 0);

            ZD_SetObjectXY(nID, x, y, TRUE);
            ProcessMessage(WM_HSCROLL, (WPARAM) MAKLNG(SB_THUMBPOSITION, gD.setAudioLevel), nID);
            gP.bin.audiolevel = gD.setAudioLevel;

        }
    }
}

static void RetractArrow() {
    long x = 0, y = 0, w = 0, h = 0;
    ZD_GetObjectXY(ID_LARROW, x, y);
    if (x > -1) {
        RECT r; GetClientRect(gP.hGDImage, &r);
        ZD_GetObjectBound (ID_LARROW, w, h);
        ZD_SetObjectXY(ID_LARROW, -w + 3, y, FALSE);
        ZD_SetObjectXY(ID_RARROW, rWidth(r) - 3, y, TRUE);
    }
}

static WCHAR* PluginPath() {
    static WCHAR zpath[MAX_PATH];
    static long done;
    if (done == 0) {
        ClearMemory(zpath, sizeof(zpath));
        Path_Combine(zpath, EXEpath(), L"BBplugin\\");
        done = -1;
    }
    return zpath;
}

static long DetectPlugin() {
    HANDLE WashFind = 0, hFind = 0;
    long nSelected = 0, nCount = 0, nRet = 0;
    WIN32_FIND_DATA fd; ClearMemory(&fd, sizeof(fd));
    WCHAR zTmp[MAX_PATH]; ClearMemory(zTmp, sizeof(zTmp));

    if (gP.hPluginlist == 0) {
        gP.hPluginlist = CreateWindowEx(0, L"LISTBOX", L"", WS_CHILD, 0, 0, 0, 0, gP.hMain, (HMENU)IDC_PLUGINLIST, GetModuleHandle(NULL), NULL);
    } else {
        List_DeleteAll(gP.hPluginlist);
    }

    Path_Combine(zTmp, PluginPath(), L"bbp_*.dll");
    hFind = FindFirstFile(zTmp, &fd);
    if (hFind != INVALID_HANDLE_VALUE) {
        do {
             Path_Combine(zTmp, PluginPath(), fd.cFileName);
             if (List_Add(gP.hPluginlist, zTmp) > -1) {
                 nCount += 1;
                 if (_wcsicmp(gP.bin.lastplugin, zTmp) == 0) nSelected = nCount;
             }
        } while (FindNextFile(hFind, &fd) != 0);
        FindClose(hFind);
    }
    nSelected = max(nSelected, 1);
    if (nCount > 0) {
        List_SelectPlus(gP.hPluginlist, nSelected);
        BBP_AttachPlugin(List_GetText(gP.hPluginlist, List_GetCursel(gP.hPluginlist)));
    }
    return nRet;
}

static void ChangePlugin(IN long nID) {
     WCHAR szFile[MAX_PATH];
     ClearMemory(szFile, sizeof(szFile));
     long nItem = List_GetCursel(gP.hPluginlist);
     if (nID == ID_LARROW) {
         if (nItem > 1) {
             nItem -= 1;
         } else {
             nItem = List_Count(gP.hPluginlist);
         }
     } else {
         if (nItem < List_Count(gP.hPluginlist)) {
             nItem += 1;
         } else {
             nItem = 1;
         }
     }
     List_SelectPlus(gP.hPluginlist, nItem);
     BBP_AttachPlugin(List_GetText(gP.hPluginlist, List_GetCursel(gP.hPluginlist)));
     MarqueeUpdate();
}

static LRESULT CALLBACK GDImageCallBack(IN HWND hWnd, IN UINT uMsg, IN WPARAM wParam, IN LPARAM lParam) {
    LRESULT nRet = 0; // Do not stop the event processing in GDImage.
    WCHAR szLabel[64]; ClearMemory(szLabel, sizeof(szLabel));
    WCHAR szFile[MAX_PATH]; ClearMemory(szFile, sizeof(szFile));
    POINT p; ClearMemory(&p, sizeof(p));
    RECT r; ClearMemory(&r, sizeof(r));
    long x = 0, y = 0, w = 0, h = 0;
    float rPercent = 0.0;

    long nID = ZI_MouseOverObjectID();

    switch (uMsg) {
        case WM_LBUTTONDOWN:
        case WM_LBUTTONDBLCLK:
            SetForegroundWindow(gP.hMain); // Avoid changing the main caption color when the GDImage layered window got the focus
            p.x = LOINT(lParam); p.y = HIINT(lParam);
            gD.focusID = 0;
            if (IsSLIDER_HORZ(nID)) {
                MoveThumbHorz(nID, p.x, ID_THUMB_HORZ, ID_SLIDER_HORZ, ID_PROGRESS_HORZ);
            } else if (IsPUSHBUTTON(nID)) {
                if (ZD_GetObjectFrameToUse(nID) != 2) {
                    ZD_SetObjectFrameToUse(nID, 2, TRUE);
                    gD.pushID = nID;
                }
            } else if (nID == ID_SEEKTRACK) {
                GetWindowRect(hWnd, &r);
               // if ((p.y > 120) && (p.y < rHeight(r) - 40)) {
ZD_GetObjectXY(ID_SEEKTHUMB, x, y);
ZD_SetObjectXY(ID_SEEKTHUMB, p.x - 7, y, FALSE);
                    rPercent = min(p.x / (float)(rWidth(r) / 100.0f), 99.9f);
                    BASS_ChannelSetPosition(gB.channel, QWORD ((gB.medialength / 100.0f) * rPercent), 0);
               //}
            }
            break;
        case WM_LBUTTONUP:
            if (IsPUSHBUTTON(nID)) {
                if (gD.pushID) {
                    ZD_SetObjectFrameToUse(gD.pushID, 3, TRUE);
                    gD.overID = gD.pushID;
                    gD.pushID = 0;
                    ProcessMessage(WM_COMMAND, MAKLNG(gD.overID, BN_CLICKED), (LPARAM)gP.hGDImage);
                }
            } else if (nID > ID_ARROW) {
                ChangePlugin(nID);
            }
            break;
        case WM_MOUSEMOVE:
            if (nID < ID_ARROW) {
                RetractArrow();
                if (ZI_IsLButtonDown()) { // Left mouse button is down
                    p.x = LOINT(lParam); p.y = HIINT(lParam);
                    if (IsSLIDER_HORZ(nID)) MoveThumbHorz(nID, p.x, ID_THUMB_HORZ, ID_SLIDER_HORZ, ID_PROGRESS_HORZ);
                } else {
                    wcscopy(szLabel, ZD_GetObjectImageLabel(nID));
                    if (my_wcsicmp(szLabel, gP.wasLabel) != 0) ZI_SetToolTipText(gP.hGDImage, szLabel);
                    wcscopy(gP.wasLabel, szLabel);
                }
            } else {
                ZD_GetObjectXY(ID_LARROW, x, y);
                if (x < 0) {
                    GetClientRect(gP.hGDImage, &r);
                    ZD_GetObjectBound (ID_LARROW, w, h);
                    ZD_SetObjectXY(ID_LARROW, 0, y, FALSE);
                    ZD_SetObjectXY(ID_RARROW, rWidth(r) - w, y, TRUE);
                }
            }
            break;
        default:
            break;
    }
    return nRet;
}

static void GDImageEnableEvents() {
    // We use a callback to monitor the GDImage control messages
    LONG_PTR lpCallBack = (LONG_PTR) GDImageCallBack;
    ZI_EventMessageEx(gP.hGDImage, lpCallBack, WM_LBUTTONDOWN, TRUE);
    ZI_EventMessageEx(gP.hGDImage, lpCallBack, WM_LBUTTONUP, TRUE);
    ZI_EventMessageEx(gP.hGDImage, lpCallBack, WM_MOUSEMOVE, TRUE);
}

static void SetThumbHorz(IN long nID, IN long THUMB_HORZ, IN long SLIDER_HORZ, IN long PROGRESS_HORZ, IN long nPercent) {
    long nX, nY, x, y, w, h, nW, nHalf;
    if (nID < THUMB_HORZ) nID += THUMB_HORZ - SLIDER_HORZ;
    ZD_GetObjectXY(nID - (THUMB_HORZ - SLIDER_HORZ), nX, y);
    ZD_GetObjectBound(nID - (THUMB_HORZ - SLIDER_HORZ), nW, h);
    ZD_GetObjectBound(nID, w, h);
    ZD_GetObjectXY(nID, x, y);
    nHalf = w / 2;
    nPercent = 100 - nPercent;
    nX = (nX + nW - 2) - ((nPercent * (nW - 5)) / 100) - nHalf;
    if (nX != x) {
        x = nX;
        ZD_GetObjectBound(PROGRESS_HORZ, w, h);
        ZD_GetObjectXY(PROGRESS_HORZ, nX, nY);
        ZD_DrawRectangleToCtrl(gP.hGDImage, nX + 1, nY + 1, x + nHalf - (nX + 1), h - 1, HIGHLIGHT, 0, PROGRESS_HORZ, ZS_VISIBLE, ZD_DRAW_FILLED, 0);
        ZD_SetObjectXY(nID, x, y, IsWindowVisible(gP.hMain));
    }
}

static void SetAudioPercentage(IN long nID, IN long nPercent) {
    nPercent = min(max(nPercent, 0), 100);
    SetThumbHorz(nID, ID_THUMB_HORZ, ID_SLIDER_HORZ, ID_PROGRESS_HORZ, nPercent);
}

static void CircularProgressBar(IN long x, IN long y, IN DWORD ColorARGB, IN WCHAR* szLabel) {
    BYTE nA, nR, nG, nB;
    zSplitColorARGB(ColorARGB, nA, nR, nG, nB);
    long nID = ID_PIE_CROWN;
    ZD_DrawArcToCtrl(gP.hGDImage, x+5, y+5, 96-5, 96-5, ZD_ARGB(28, nR,nG,nB), 5*2, nID, ZS_VISIBLE, -90, 360);
    ZD_SetObjectLocked(nID, TRUE);
    ZD_SetObjectImageLabel(nID, szLabel);

    nID = ID_PIE_OUT;
    ZD_DrawPieToCtrl(gP.hGDImage, x, y, 96, 96, ZD_ARGB(nA, nR,nG,nB), 0, nID, ZS_VISIBLE, ZD_DRAW_FILLED, 0, -90, 0);
    ZD_SetObjectLocked(nID, TRUE);
    ZD_SetObjectImageLabel(nID, szLabel);

    nID = ID_PIE_IN;
    ZD_DrawPieToCtrl(gP.hGDImage, x+10, y+10, 86, 86, ZD_ARGB(227, 0,0,0), 0, nID, ZS_VISIBLE, ZD_DRAW_FILLED, 0, -90, 360);
    ZD_SetObjectLocked(nID, TRUE);
    ZD_SetObjectImageLabel(nID, szLabel);

    x += 15; y += 18;
    nID = ID_PIE_TEXT;
    long nFontSize = 18;
    ZD_DrawTextToCtrlEx(gP.hGDImage, L"", x, y, 50, 40, OPAQUEIMAGE, L"TREBUCHET MS", nFontSize, nID, ZS_VISIBLE, -1, StringAlignmentCenter);
    ZD_SetObjectLocked(nID, TRUE);
    ZD_SetObjectImageLabel(nID, szLabel);
}

static void CreateGDImageControls() {
    long nID = 0, x = 0, y = 0, w = 0, h = 0;

    HBITMAP hBitmap = ZI_CreateBitmapFromFile(Resource(L"larrow.png"), w, h);
    if (hBitmap) {
        nID = ID_LARROW;
        x = -w + 3; 
        y = (ClientH - h) / 2;
        ZD_DrawBitmapToCtrl(gP.hGDImage, x, y, hBitmap, ZD_ColorARGB(255, 0), nID, ZS_VISIBLE);
        ZD_SetObjectLocked(nID, TRUE);
    }

    hBitmap = ZI_CreateBitmapFromFile(Resource(L"rarrow.png"), w, h);
    if (hBitmap) {
        nID = ID_RARROW;
        x = ClientW - 3;
        y = (ClientH - h) / 2;;
        ZD_DrawBitmapToCtrl(gP.hGDImage, x, y, hBitmap, ZD_ColorARGB(255, 0), nID, ZS_VISIBLE);
        ZD_SetObjectLocked(nID, TRUE);
    }

    // Create the bottom of the horizontal slider
    hBitmap = ZI_CreateBitmapFromFile(Resource(L"sliderH_bottom.png"), w, h);
    x = (ClientW - w) / 2, y = ClientH - (h + 3);
    if (hBitmap) {
        nID = ID_BACKGROUND;
        ZD_DrawBitmapToCtrl(gP.hGDImage, x, y, hBitmap, OPAQUEIMAGE, nID, ZS_VISIBLE);
        ZD_SetObjectHidden(nID, 2); // 2 = Transparent, its ID won't be detected and it couldn't be moved.
        ZD_SetObjectAnchorMode(nID, ANCHOR_CENTER_HORZ_BOTTOM);
    }

    // Horizontal slider
    hBitmap = ZI_CreateBitmapFromFile(Resource(L"slideH.png"), w, h);
    if (hBitmap) {
        HBITMAP hThumb = ZI_CreateBitmapFromFile(Resource(L"thumbH.png"), w, h);
        if (hThumb) {
            CreateSliderHorz(hBitmap, hThumb, ID_SLIDER_HORZ, ID_THUMB_HORZ, ID_PROGRESS_HORZ, x + 48, y);
            ZD_SetObjectImageLabel(ID_SLIDER_HORZ, L"Audio volume"); // We use this to display the tooltip text
        }
    }

    // Button PAUSE
    hBitmap = ZI_CreateBitmapFromFile(Resource(L"pause_btn.png"), w, h);
    if (hBitmap) {
        nID = ID_BTN_PAUSE;
        ZD_DrawBitmapToCtrl(gP.hGDImage, x, y, hBitmap, OPAQUEIMAGE, nID, ZS_VISIBLE);
        ZD_SetObjectFrameCount(nID, 3);
        ZD_SetObjectFrameToUse(nID, 1, FALSE);
        ZD_SetObjectAnchorMode(nID, ANCHOR_CENTER_HORZ_BOTTOM);
        ZD_SetObjectLocked(nID, TRUE);
        ZD_SetObjectImageLabel(nID, L"Pause"); // We use this to display the tooltip text
    }
    // Button PLAY
    hBitmap = ZI_CreateBitmapFromFile(Resource(L"play_btn.png"), w, h);
    if (hBitmap) {
        nID = ID_BTN_PLAY;
        ZD_DrawBitmapToCtrl(gP.hGDImage, x, y, hBitmap, OPAQUEIMAGE, nID, ZS_HIDDEN);
        ZD_SetObjectFrameCount(nID, 3);
        ZD_SetObjectFrameToUse(nID, 1, FALSE);
        ZD_SetObjectAnchorMode(nID, ANCHOR_CENTER_HORZ_BOTTOM);
        ZD_SetObjectLocked(nID, TRUE);
        ZD_SetObjectImageLabel(nID, L"Play"); // We use this to display the tooltip text
    }

    // Circular progress bar
    x = 20; y = 20;
    CircularProgressBar(x, y, HIGHLIGHT, L"Progress gauge");

    nID = ID_SEEKTRACK;
    BYTE nA = 0, nR = 0, nG = 0, nB = 0;
    zSplitColorARGB(HIGHLIGHT, nA, nR, nG, nB);
    ZD_DrawRectangleToCtrl(gP.hGDImage, 0, ClientH - 54, 3840, 14, ZD_ARGB(16,  nR, nG, nB), 0, nID, ZS_VISIBLE, ZD_DRAW_FILLED, 0);
    ZD_SetObjectLocked(nID, TRUE);
    ZD_SetObjectAnchorMode(nID, ANCHOR_BOTTOM);
    ZD_SetObjectImageLabel(nID, L"Audio seek tracker"); // We use this to display the tooltip text

    nID = ID_SEEKTHUMB;
    ZD_DrawRectangleToCtrl(gP.hGDImage, -7, ClientH - 54, 14, 14, ZD_ARGB(32,  255, 255, 255), 0, nID, ZS_VISIBLE, ZD_DRAW_OUTLINE, 0);
    ZD_SetObjectLocked(nID, TRUE);
    ZD_SetObjectAnchorMode(nID, ANCHOR_BOTTOM);
    ZD_SetObjectImageLabel(nID, L"Audio seek thumb"); // We use this to display the tooltip text

    // Default audio level
    SetAudioPercentage(ID_THUMB_HORZ, gD.setAudioLevel);

    // Add tooltip support
    ZI_CreateToolTip(gP.hGDImage, L"");

    // PreProcess GDImage events
    GDImageEnableEvents();
}

int WINAPI wWinMain(_In_ HINSTANCE hInstance, _In_opt_ HINSTANCE hPrevInstance, _In_ LPWSTR lpCmdLine, _In_ int nShowCmd) {

    UNREFERENCED_PARAMETER(hPrevInstance);
    UNREFERENCED_PARAMETER(lpCmdLine);

    long nRet = 0;

    WCHAR szClass[] = L"ZMATRIX";

    // Detect process without using a Mutex
    HWND hFound = FindWindow(szClass, NULL);
    if (hFound) {
        if (IsIconic(hFound)) ShowWindow(hFound, SW_RESTORE);
        SetForegroundWindow(hFound);
        return nRet;
    }

    // Register our main window class
    WNDCLASSEX wcx; ClearMemory(&wcx, sizeof(wcx));
    wcx.cbSize = sizeof(wcx);
    long IsInitialized = GetClassInfoEx(hInstance, szClass, &wcx);
    if (!IsInitialized) {
        wcx.style         = 0;//CS_HREDRAW | CS_VREDRAW;
        wcx.lpfnWndProc   = WndProc;
        wcx.cbClsExtra    = 0;
        wcx.cbWndExtra    = 0;
        wcx.hInstance     = hInstance;
        wcx.hIcon         = LoadIcon(hInstance, MAKEINTRESOURCE(IDI_ICON1));
        wcx.hCursor       = LoadCursor(NULL, IDC_ARROW);
        wcx.hbrBackground = NULL;//(HBRUSH)(COLOR_WINDOW + 1);
        wcx.lpszMenuName  = NULL;
        wcx.lpszClassName = szClass;
        wcx.hIconSm       = wcx.hIcon;
        if (RegisterClassEx(&wcx)) { IsInitialized = -1; }
    }
    if (IsInitialized) {

        // Load the GDImage.dll
        if (RegisterGDImageClass() == 0) { // If it fails to register then return ERROR = 99
            return 99;
        } else {
            ZI_LoadDLL(L"");
        }

        // Window Extended Style
        DWORD dwExStyle = WS_EX_ACCEPTFILES;
        DWORD dwStyle = WS_POPUP | WS_CAPTION | WS_SYSMENU | WS_THICKFRAME | WS_MINIMIZEBOX | WS_MAXIMIZEBOX;

        RECT rc; SetRect(&rc, 0, 0, ClientW, ClientH);
        AdjustWindowRectEx(&rc, dwStyle, FALSE, dwExStyle);    // Adjust Window To True Requested Size

        gP.MinTrackSizeW = rWidth(rc);
        gP.MinTrackSizeH = rHeight(rc);
        long x = max((GetSystemMetrics(SM_CXSCREEN) - gP.MinTrackSizeW) / 2, 0);
        long y = max((GetSystemMetrics(SM_CYSCREEN) - gP.MinTrackSizeH) / 2, 0);

        // Create the MAIN window
        gP.hMain = CreateWindowEx(dwExStyle,              // Extended Style For The Window
                             szClass,                     // Class Name
                             L"BassBox plugin 64-bit",    // Window Title
                             dwStyle,                     // Defined Window Style
                             x, y,                        // Window Position
                             gP.MinTrackSizeW,            // Calculate Window Width 
                             gP.MinTrackSizeH,            // Calculate Window Height
                             NULL,                        // No Parent Window
                             NULL,                        // No Menu
                             hInstance,                   // Instance
                             NULL);                       // Dont Pass Anything To WM_CREATE

        if (gP.hMain) {

            // Create a WGL GDImage control supporting OpenGL extensions.
            dwExStyle = 0; dwStyle = WS_CHILD | WS_VISIBLE;
            gP.hGL = WGL_CreateWindow(dwExStyle, dwStyle, 0, 0, ClientW, ClientH, gP.hMain, (HMENU) IDC_GLCTRL);

            if (gP.hGL) {
                // Start skinning the MAIN window;
                if (skInitEngine(Resource(L"Plugin.sks"), 0)) {
                    skSkinWindow(gP.hMain, L"");

                    // When using skSkinWindow we must update MinTrackSize once done
                    GetWindowRect(gP.hMain, &rc);
                    gP.MinTrackSizeW = rWidth(rc);
                    gP.MinTrackSizeH = rHeight(rc);

                    // Anchor the GDImage WGL control (to let it be resizable).
                    skSetAnchorCtrl(gP.hGL, ANCHOR_HEIGHT_WIDTH);

                    LoadConfig();
                    DetectPlugin();

                    // Create a popup GDImage overlay window
                    gP.hGDImage = CreateWindowEx(WS_EX_TOOLWINDOW | WS_EX_NOACTIVATE, GDImageClassName, NULL, WS_POPUP | WS_VISIBLE, x + skGetSystemMetrics(SK_CXFRAMELEFT), y + skGetSystemMetrics(SK_CYFRAMETOP), ClientW, ClientH, gP.hMain, 0, hInstance, NULL);
                    if (IsWindow(gP.hGDImage)) {
                        ZI_DwmEnable(gP.hGDImage);
                        CreateGDImageControls();
                    }

                    // Display the plugin copyright
                    ZD_DrawTextToCtrl(gP.hGDImage, gP.plugintitle, ClientW + 10, 0, ZD_ARGB(255, 192,192,192), L"ARIAL", 13, ID_COPYRIGHT, ZS_VISIBLE, 0, 0);
                    ZD_SetObjectHidden(ID_COPYRIGHT, 2); // 2 = Transparent, its ID won't be detected and it couldn't be moved.

                    // Start playing audio using Bass.dll.
                    wcscopy(gB.audiofile, Resource(L"Tracker.mod"));
                    BassChannelPlay(gP.hMain);
                    if (gB.channel) BBP_NewSound();
                    // Update the marquee with the audio file name.
                    MarqueeUpdate();

                }

                // Show the main window
                ShowWindow(gP.hMain, nShowCmd);
                SetForegroundWindow(gP.hMain); // Slightly Higher Priority

                // Main message pump.
                MSG msg; ClearMemory(&msg, sizeof(msg));
                SetTimer(gP.hMain, IDC_TIMER, 0, 0); // 66 FPS
                while (GetMessage(&msg, NULL, 0, 0)) {
                    TranslateMessage(&msg);
                    DispatchMessage(&msg);
                }
                KillTimer(gP.hMain, IDC_TIMER);
                nRet = (long) msg.wParam;

                //// Cleanup texture from memory
                //ZI_DestroyTexture(&gP.tex[0], TEX_MAX);

            }
        }
    }
    return nRet;
}
